
from Planet import Planet
from MsgEvent import MsgEvent
import string
import math
import random

DEF_PLYR_INDUST = 10
DEF_PLYR_SHIPS = 100

DEF_FIELD_SIDE = 20
DEF_NUM_DIMENSIONS = 2
DEF_NUM_PLANETS = 37
DEF_MAX_INDUST = 6
DEF_MAX_RANDSHIPS = 5

MIN_FIELD_DIM = 10

class Game:
	def __init__(self, server, id, owner):
		self.events = EventQueue(self)
		self.field = StarField(self, self.events, server)
		self.closed = 0
		self.started = 0
		self.players = []
		self.busy_players = []
		self.ownerpl = owner
		self.server = server
		self.year = 0.0
		self.id = id
		self.name = ""
		self.commands = {	'JOIN': self.recv_join, \
					'PFLD': self.recv_field, \
					'PAFD': self.recv_field, \
					'PPLN': self.recv_field, \
					'PLAY': self.recv_state, \
					'PCLO': self.recv_state, \
					'MESG': self.recv_message, \
					'MOVE': self.recv_player_move, \
					'DONE': self.recv_player_done, \
					'SUIC': self.recv_player_suicide, \
					'QUIT': self.recv_player_suicide }

	def get_selfaseventlist(self):
		if self.closed:
			return []
		evl = [MsgEvent('GAME', game=self, player=self.ownerpl, \
				quant=len(self.players),  name=self.name, \
				desc=self.plyrlistasstring())]
		map(evl.append, self.field.get_selfaseventlist())
		return evl

	def plyrlistasstring(self):
		plnames = []
		for p in self.players:
			plnames.append(p.name)
		return "(%s)" % string.join(plnames)

	def send_broadcast(self, event):
		for p in self.players:
			p.send_event(event)

	def reg_player(self, player):
		if self.closed:
			raise GameClosed
		self.players.append(player)
		self.send_broadcast(player.get_selfasevent('GPLR'))

	def unreg_player(self, player):
		self.send_broadcast(player.get_selfasevent('QUIP'))
		self.mark_player_done(player)
		self.players.remove(player)
		player.unreg_game(self)
		self.check_winner()

	def check_winner(self):
		playing = filter(lambda p: p.is_alive(self), self.players)
		if len(playing) > 1: return
		if not self.started and self.ownerpl in playing: return
		if len(playing) == 1:
			winner = playing[0]
			event = winner.get_selfasevent('WINN')
			self.send_broadcast(event)
		self.end_game()

	def end_game(self):
		for p in self.players:
			self.unreg_player(p)
		self.server.unreg_game(self.id)

	def start_year(self, year):
		self.set_year(year)
		self.field.yearly(year)
		self.busy_players = filter(lambda p: p.is_alive(self),
				self.players)
		self.field.send_planets()
		self.send_broadcast(MsgEvent('BRND', quant=self.year))

	def start_playing(self):
		self.started = 1
		self.set_year(0.0)
		self.send_initinfo(self.send_broadcast)
		self.start_year(1.0)

	def send_initinfo(self, sfunc):
		map(sfunc, self.field.get_selfaseventlist())
		map(sfunc, self.field.get_planetsaseventlist())

	def done_players(self):
		return filter(lambda a: a not in self.busy_players,
				self.players)

	def set_year(self, year):
		self.year = year
		self.send_broadcast(MsgEvent('TIME', quant=year))

	def get_year(self):
		return self.year

	def advance_year(self):
		nxyear = self.year + 1.0
		self.events.launch_up_to(nxyear)
		self.start_year(nxyear)

	# Here beginneth the command handler section
	def recv_event(self, event):
		try:
			command = self.commands[event.evtype]
		except KeyError:
			self.server.recv_event(event)
			return
		command(event)

	def recv_join(self, event):
		if self.closed:
			event.player.send_event(MsgEvent('ERRP', name=event.evtype, \
							 desc="The game is closed."))
			return
		self.field.recv_event(event)
		event.player.reg_game(self)
		event.player.send_event(event.player.get_selfasevent('JNOK'))
		event.player.send_event(MsgEvent('OKAY', name=event.evtype))
		event.player.send_event(MsgEvent('GOWN', player=self.ownerpl, game=self))
		pl = []
		for p in self.players:
			pl.append(p.get_selfasevent('GPLR'))
		self.reg_player(event.player)
		map(self.server.send_listenerbc, self.get_selfaseventlist())
		map(event.player.send_event, pl)
		if self.started:
			self.send_initinfo(event.player.send_event)

	def recv_field(self, event):
		if event.player != self.ownerpl:
			event.player.send_event(MsgEvent('ERRP', name=event.evtype, \
							 desc="You don't own this game."))
			return
		self.field.recv_event(event)
		map(self.server.send_listenerbc, self.get_selfaseventlist())

	def recv_state(self, event):
		if event.player != self.ownerpl:
			event.player.send_event(MsgEvent('ERRP', name=event.evtype, \
							 desc="You don't own this game."))
			return
		if event.evtype == 'PCLO':
			self.closed = 1
			event.player.send_event(MsgEvent('OKAY', name=event.evtype))
			return
		if self.started:
			event.player.send_event(MsgEvent('ERRP', name=event.evtype, \
							 desc="The game is already started."))
		event.player.send_event(MsgEvent('OKAY', name=event.evtype))
		self.start_playing()

	def recv_player_move(self, event):
		if not self.started:
			event.player.send_event(MsgEvent('ERRP', name=event.evtype, \
							 desc="The game has not been started."))
			return
		self.field.recv_event(event)

	def recv_player_done(self, event):
		if not self.started:
			event.player.send_event(MsgEvent('ERRP', name=event.evtype, \
							 desc="The game has not been started."))
			return
		if not event.player in self.busy_players:
			event.player.send_event(MsgEvent('ERRP', name=event.evtype, \
							 desc="You're done for this turn!"))
			return
		event.player.send_event(MsgEvent('OKAY', name=event.evtype))
		self.mark_player_done(event.player)

	def mark_player_done(self, player):
		if player in self.busy_players:
			self.busy_players.remove(player)
			ev = player.get_selfasevent('PDON')
			for p in self.done_players():
				if p != player: p.send_event(ev)
				player.send_event(p.get_selfasevent('PDON'))
		if len(self.busy_players) < 1:
			self.send_broadcast(MsgEvent('ERND'))
			self.advance_year()

	def recv_player_suicide(self, event):
		ev = MsgEvent('PKIL', player=event.player, name=event.name, desc=event.desc)
		self.send_broadcast(ev)
		self.mark_player_done(event.player)
		self.check_winner()

	def recv_message(self, event):
		if event.tg_player != None:
			self.server.recv_event(event)
			return
		self.send_broadcast(MsgEvent('BMSG', player=event.player, desc=event.desc))

class EventQueue:
	def __init__(self, game):
		self.game = game
		self.qu = []

	def schedule_move(self, event):
		def add(a, b): return a+b
		def rr(a, b): return (a-b)*(a-b)
		dd = reduce(add, map(rr, event.planet.get_location(),
				     event.tg_planet.get_location()))
		traveltime = math.sqrt(dd)
		eta = self.game.get_year() + traveltime
		self.qu.append((eta, event))
		event.player.bkeep_move(event)
		return eta

	def launch_up_to(self, year):
		self.qu.sort()
		try:
			while self.qu[0][0] <= year:
				y = self.qu[0][0]
				self.game.set_year(y)
				event = self.qu[0][1]
				event.player.ubkeep_move(event)
				event.tg_planet.recv_arrive(event, y)
				del self.qu[0]
		except IndexError:		# No more events
			pass

class StarField:
	def __init__(self, game, eventq, server):
		self.game = game
		self.server = server
		self.eventq = eventq
		self.dimensions = tuple([DEF_FIELD_SIDE] * DEF_NUM_DIMENSIONS)
		self.planets = {}
		self.plids = {}
		self.nxplid = 0
		self.spy_aging = 10.0
		self.name = ""
		self.commands = {	'MOVE' : self.recv_player_move, \
					'JOIN' : self.recv_player_join, \
					'PFLD' : self.recv_winch, \
					'PPLN' : self.recv_planets }

	def add_planets(self, amount):
		while amount > 0:
			self.found_planet(self.server.get_player(0), \
					  random.randint(0, DEF_MAX_INDUST), \
					  random.randint(0, DEF_MAX_RANDSHIPS))
			amount = amount - 1

	def found_planet(self, owner, indust, ships):
		planet = Planet(self, self.nxplid, owner, indust, ships)
		if owner.id != 0:
			self.game.send_broadcast(planet.get_selfasevent('PLNT'))
		self.plids[self.nxplid] = planet
		self.nxplid = self.nxplid + 1

	def get_planet(self, id):
		return self.plids[id]

	def get_dimensions(self):
		return self.dimensions

	def get_planetat(self, loc):
		if self.planets.has_key(loc):
			return self.planets[loc]
		else:
			return None

	def reg_planet(self, planet):
		loc = planet.get_location()
		if loc == None:
			return
		if self.get_planetat(loc):
			raise PlanetCollision
		self.planets[loc] = planet

	def get_selfaseventlist(self):
		return [MsgEvent('GFLD', name=self.name, loc=self.dimensions, \
				 game=self.game, quant=len(self.plids))]

	def get_planetsaseventlist(self):
		evl = []
		for p in self.plids.values():
			evl.append(p.get_selfasevent('PLNT'))
		return evl

	def yearly(self, year):
		for p in self.plids.values():
			p.yearly(year)

	def get_prob_factor(self, event):
		return 3.0

	def send_planets(self):
		for p in self.plids.values():
			p.send_selfbroadcast('PIND')
			p.send_selfbroadcast('PSHP')

	# Here beginneth the command handler section
	def recv_event(self, event):
		try:
			command = self.commands[event.evtype]
		except KeyError:
			ev = MsgEvent('ERRS', name=event.evtype, \
				      desc="This command has not been implemented.")
			event.player.send_event(ev)
			return
		command(event)
			
	def recv_player_join(self, event):
		self.found_planet(event.player, DEF_PLYR_INDUST, DEF_PLYR_SHIPS)

	def recv_player_move(self, event):
		if not event.planet.check_move(event):
			return
		eta = self.eventq.schedule_move(event)
		event.planet.record_move(event, eta)

	def recv_winch(self, event):
		def bound(num):
			if num < MIN_FIELD_DIM:
				return MIN_FIELD_DIM
			return num
		self.dimensions = map(bound, event.loc)
		self.name = event.name
		self.planets = {}
		for p in self.plids.values():
			p.winch()
		event.player.send_event(MsgEvent('OKAY', name=event.evtype))

	def recv_planets(self, event):
		self.add_planets(event.quant)

