
from MsgEvent import MsgEvent
from Selecting.SelectManager import ChannelHandler
import string

protver = "0.3"
serverid = "GalaxIEs Galax Internet Experience Server"

class NullProto:
	def __init__(self, player):
		pass
	def send_event(self, event):
		pass
	def got_game(self, game):
		pass

class UndefProto (ChannelHandler):
	def __init__(self, player):
		self.player = player
		self.conn = None
		self.open = 1
		player.regproto(self)
		self.protocols = {	'0.2' : Ver021Proto, \
					'0.21' : Ver021Proto, \
					'0.22' : Ver021Proto, \
					'0.23' : Ver023Proto, \
					'0.3' : Ver03Proto }

	def reg_conn(self, conn):
		self.conn = conn
		conn.reg_handler(self)
		self.send("GSRV %s %s" % (protver, serverid))

	def receive(self, data):
		cmd = string.split(data)
		try:
			if string.upper(cmd[0]) != 'GCLT':
				self.send("ERRC Please use GCLT to make a handshake")
				return
			proto = self.protocols[cmd[1]]
		except IndexError:
			self.send("ERRC unsupported protocol version")
			return
		newproto = proto(self.player, string.join(cmd[2:]))
		newproto.reg_conn(self.conn)
		self.player.handshake_afterhook()

	def send(self, data):
		if self.open:
			self.conn.write(data)

	def recv_event(self, event):
		if event.evtype == 'QUIT': self.open = 0
		self.player.recv_event(event)

	def send_event(self, event):
		pass

	def cleanup(self):
		self.recv_event(MsgEvent('QUIT', desc="Connection reset"))

class Ver021Proto(UndefProto):
	def __init__(self, player, desc):
		self.player = player
		self.time = -1.0
		self.sendgame = 1
		self.open = 1
		self.pl_cache = {}
		self.mrg = None
		self.desc = desc
		player.regproto(self)
		self.cmdhandler = {	'JOIN': self.cmd_join, \
					'PLAY': self.cmd_play, \
					'MOVE': self.cmd_move, \
					'DONE': self.cmd_done, \
					'QUIT': self.cmd_quit }
		self.events = {		'BGML': self.send_clok, \
					'EGML': self.send_slct, \
					'JNOK': self.send_jnok, \
					'GPLR': self.send_gplr, \
					'QUIP': self.send_kill, \
					'WINN': self.send_winn, \
					'GAME': self.maybe_send_game, \
					'BRND': self.send_year_and_pl, \
					'TIME': self.set_time, \
					'GFLD': self.maybe_send_gfld, \
					'GOWN': self.make_planets, \
					'PLNT': self.cache_pl_info, \
					'PIND': self.cache_pl_indust, \
					'PSHP': self.cache_pl_ships, \
					'MVOK': self.send_mvok, \
					'BATT': self.send_batt, \
					'ATTK': self.send_attk, \
					'BWIN': self.send_bwin, \
					'FORT': self.send_fort, \
					'ERRS': self.send_error, \
					'ERRP': self.send_error, \
					'ERRC': self.send_error }

	def reg_conn(self, conn):
		self.conn = conn
		conn.reg_handler(self)

	def receive(self, data):
		cmd = string.split(data)
		try:
			args = cmd[1:]
		except IndexError:
			args = []
		try:
			handler = self.cmdhandler[string.upper(cmd[0])]
		except (KeyError, IndexError):
			self.send("ERRC unknown command %s" % data)
			return
		handler(args)

	def got_game(self, game):
		self.mrg = game

	#Incoming
	def cmd_join(self, args):
		try:
			gameid = string.atoi(args[0])
			game = server.get_or_found_game(gameid, self.player)
			event = MsgEvent('JOIN', player=self.player, \
					 game=game, name=args[1], \
					 desc=self.desc)
		except (ValueError, IndexError):
			self.send("ERRC Please specify game id and player name")
			return
		self.player.recv_event(event)

	def cmd_play(self, args):
		if self.mrg == None:
			self.send("ERRC Please join a game first")
			return
		self.player.recv_event(MsgEvent('PLAY', player=self.player, game=self.mrg))
		self.player.recv_event(MsgEvent('PCLO', player=self.player, game=self.mrg))

	def cmd_move(self, args):
		if self.mrg == None:
			self.send("ERRC Please join a game first")
			return
		try:
			[fr, to, num] = map(string.atoi, args)
			frpl = self.mrg.field.get_planet(fr)
			topl = self.mrg.field.get_planet(to)
			ev = MsgEvent('MOVE', player=self.player, quant=num, \
				      planet=frpl, tg_planet=topl, game=self.mrg)
		except ValueError:
			self.send("ERRC Please specify from, to, amount")
			return
		except KeyError:
			self.send("ERRC No such planet")
			return
		self.player.recv_event(ev)

	def cmd_done(self, args):
		if self.mrg == None:
			self.send("ERRC Please join a game first")
			return
		self.player.recv_event(MsgEvent('DONE', player=self.player, game=self.mrg))

	def cmd_quit(self, args):
		self.player.recv_event(MsgEvent('QUIT', player=self.player))

	#Outgoing
	def make_planets(self, event):
		if event.player == self.player:
			self.player.recv_event(MsgEvent('PPLN', player=self.player, game=event.game, quant=50))

	def send_event(self, event):
		try:
			handler = self.events[event.evtype]
		except KeyError:
			return
		handler(event)

	def cache_pl_info(self, event):
		try:
			self.pl_cache[event.planet.id] = [event.loc[0], event.loc[1], \
							  event.player.id, -1, -1]
		except KeyError:
			self.pl_cache[event.planet.id] = [-1, -1, -1, -1, -1]
			
	def cache_pl_indust(self, event):
		try:
			self.pl_cache[event.planet.id][4] = event.quant
		except KeyError:
			self.pl_cache[event.planet.id] = [-1, -1, -1, -1, event.quant]

	def cache_pl_ships(self, event):
		try:
			self.pl_cache[event.planet.id][3] = event.quant
			self.pl_cache[event.planet.id][2] = event.player.id
		except KeyError:
			self.pl_cache[event.planet.id] = [-1, -1, event.player.id, event.quant, -1]

	def set_time(self, event):
		self.time = event.quant
	def send_clok(self, event):
		self.send("CLOK")
	def send_slct(self, event):
		self.send("SLCT")
		self.sendgame = 0
	def send_jnok(self, event):
		self.send("JNOK %d %d %s" % (event.player.game.id, event.player.id, event.name))
	def send_gplr(self, event):
		self.send("GPLR %d %s %s" % (event.player.id, event.name, event.desc))
	def send_kill(self, event):
		self.send("KILL %d" % (event.player.id,))
	def send_winn(self, event):
		self.send("WINN %d" % (event.player.id,))
	def maybe_send_game(self, event):
		if self.sendgame:
			self.send("GAME %d %d %s" % (event.game.id, event.quant, event.desc))
	def send_year_and_pl(self, event):
		self.set_time(event)
		for i in self.pl_cache.keys():
			pl_info = [i]
			map(pl_info.append, self.pl_cache[i])
			self.send("PLNT %d %d %d %d %d %d" % tuple(pl_info))
			self.pl_cache[i][3] = -1
		self.send("YEAR %.1f" % self.time)
	def maybe_send_gfld(self, event):
		if self.time != 0.0:
			return
		self.send("GFLD %d %d %d -1 0.21" % \
			  (event.loc[0], event.loc[1], event.quant))
	def send_mvok(self, event):
		self.send("MVOK %d %d %.2f" % (event.planet.id, event.quant, event.change))
	def send_batt(self, event):
		self.send("BATT %.2f %d %d %d %d %d %d" % \
			  (self.time, event.planet.id, event.player.id, \
			   event.tg_player.id, event.which, event.quant, \
			   event.change))
		if event.which:
			self.attshcache = event.change
		else:
			self.attshcache = event.quant
	def send_attk(self, event):
		self.send("ATTK %d %d %d %d" % \
			  (event.which, event.change, \
			   event.quant[1], event.quant[0]))
	def send_bwin(self, event):
		self.send("BWIN %d %d %d %d %d" % \
			  (event.which, event.player.id, event.tg_player.id, \
			   event.quant, event.planet.id))
	def send_fort(self, event):
		self.send("FORT %.1f %d %d %d %d" % \
			  (self.time, event.planet.id, event.player.id, \
			   event.quant, event.change))
	def send_error(self,event):
		if event.evtype == 'ERRC':
			err = 'ERRC'
		elif event.name == 'JOIN':
			err = 'ERRJ'
		elif event.name == 'MOVE':
			err = 'ERRM'
		else:
			err = 'ERRP'
		self.send("%s %s" % (err, event.desc))

class Ver023Proto(Ver021Proto):
	def __init__(self, player, desc):
		Ver021Proto.__init__(self, player, desc)
		self.events['PDON'] = self.send_pdon
		self.events['GOWN'] = self.send_gown
		self.events['GAME'] = self.send_game

	def send_pdon(self, event):
		self.send("PDON %d" % event.player.id)
	def send_gown(self, event):
		if event.player == self.player:
			self.player.recv_event(MsgEvent('PPLN', player=self.player, game=event.game, quant=50))
		self.send("GOWN %d %d" % (event.game.id, event.player.id))
	def send_game(self, event):
		self.send("GAME %d %d %s" % (event.game.id, event.quant, event.desc))

class Ver03Proto(UndefProto):
	def __init__(self, player, desc):
		self.player = player
		self.mrg = None
		self.desc = desc
		self.open = 1
		player.regproto(self)
		self.fieldstr = "CFLD"
		self.cmdhandler = {	'QUIT' : self.recv_quit, \
					'RSET' : self.recv_reset, \
					'JOIN' : self.recv_join, \
					'FILD' : self.recv_field, \
					'PGAM' : self.recv_game_stat, \
					'MOVE' : self.recv_move, \
					'DONE' : self.recv_done, \
					'SUIC' : self.recv_suicide, \
					'MESG' : self.recv_message }
		self.events = {		'BGML' : self.send_clok, \
					'EGML' : self.send_slct, \
					'GPLR' : self.send_gplr, \
					'PDON' : self.send_pdon, \
					'GOWN' : self.send_gown, \
					'QUIP' : self.send_pquit, \
					'WINN' : self.send_winner, \
					'GAME' : self.send_game, \
					'BRND' : self.send_year, \
					'TIME' : self.send_time, \
					'ERND' : self.send_ernd, \
					'PKIL' : self.send_psuicide, \
					'GFLD' : self.send_c_or_g_fld, \
					'PLNT' : self.send_planet, \
					'PIND' : self.send_pindust, \
					'PSHP' : self.send_pships, \
					'MVOK' : self.send_move_okay, \
					'BATT' : self.send_battle, \
					'ATTK' : self.send_attack, \
					'BWIN' : self.send_bwin, \
					'FORT' : self.send_fort, \
					'PMSG' : self.send_message, \
					'BMSG' : self.send_message, \
					'JNOK' : self.send_jnok, \
					'OKAY' : self.send_okay, \
					'ERRS' : self.send_error, \
					'ERRC' : self.send_error, \
					'ERRP' : self.send_error }

	def reg_conn(self, conn):
		self.conn = conn
		conn.reg_handler(self)

	def receive(self, data):
		cmd = string.split(data)
		args = {}
		margs = []
		try:
			for a in cmd[1:]:
				aval = string.split(a, "=")
				try:
					args[string.upper(aval[0])] = aval[1]
				except IndexError:
					margs.append(a)
		except IndexError:
			pass
		try:
			self.cmdhandler[string.upper(cmd[0])](args, margs)
		except (KeyError, IndexError):
			self.send("ERRC unknown command %s" % data)

	def got_game(self, game):
		self.mrg = game

	# Incoming
	def recv_game_stat(self, args, margs):
		try:
			gameid = string.atoi(args["G"])
			try:
				game = server.get_game(gameid)
			except KeyError:
				self.send("ERRP no such game")
				return
			if string.upper(args["TYPE"]) == 'C':
				evtype='PCLO'
			else:
				evtype='PLAY'
			event = MsgEvent(evtype, game=game, player=self.player)
			self.player.recv_event(event)
		except (ValueError, KeyError):
			self.send("ERRC Usage PGAM g=<game> type=<S|C>")

	def recv_field(self, args, margs):
		try:
			gameid = string.atoi(args["G"])
			try:
				game = server.get_game(gameid)
			except KeyError:
				self.send("ERRP no such game")
				return
			dimens = string.atoi(args["X"]), string.atoi(args["Y"])
			self.player.recv_event(MsgEvent('PFLD', player=self.player, game=game, loc=dimens, name=""))
			numpl = string.atoi(args["PL"])
			self.player.recv_event(MsgEvent('PPLN', player=self.player, game=game, quant=numpl))
		except (ValueError, KeyError):
			self.send("ERRC Usage FILD g=<game> x=<x-size> y=<y-size> pl=<#planets>")

	def recv_join(self, args, margs):
		try:
			gameid = string.atoi(args["G"])
			game = server.get_or_found_game(gameid, self.player)
			self.player.recv_event(MsgEvent('JOIN', player=self.player, game=game, \
					       name=args["NAME"], desc=self.desc))
		except (ValueError, KeyError):
			self.send("ERRC Usage JOIN g=<game> name=<name> <Your long description>")

	def recv_move(self, args, margs):
		if self.mrg == None:
			self.send("ERRP Join first")
			return
		try:
			frid = string.atoi(args["FROM"])
			toid = string.atoi(args["TO"])
			try:
				frpl = self.mrg.field.get_planet(frid)
				topl = self.mrg.field.get_planet(toid)
			except KeyError:
				self.send("ERRP no such planet")
			shcnt = string.atoi(args["S"])
			self.player.recv_event(MsgEvent('MOVE', player=self.player, game=self.mrg, \
							planet=frpl, tg_planet=topl, quant=shcnt))
		except (ValueError, KeyError):
			self.send("ERRC Usage MOVE from=<planet> to=<destination> s=<#ships>")

	def recv_done(self, args, margs):
		if self.mrg == None:
			self.send("ERRP Join first")
			return
		self.player.recv_event(MsgEvent('DONE', player=self.player, game=self.mrg))

	def recv_suicide(self, args, margs):
		if self.mrg == None:
			self.send("ERRP Quitting before starting?... <shrug>")
			return
		self.player.recv_event(MsgEvent('SUIC', player=self.player, game=self.mrg, \
						desc = "Suicide: " + string.join(margs)))

	def recv_message(self, args, margs):
		if self.mrg == None:
			self.send("ERRP Messages can only be sent to one game's players.")
			return
		self.player.recv_event(MsgEvent('MESG', player=self.player, game=self.mrg, \
						desc=string.join(margs)))

	def recv_reset(self, args, margs):
		proto = UndefProto(self.player)
		proto.reg_conn(self.conn)

	def recv_quit(self, args, margs):
		self.player.recv_event(MsgEvent('QUIT', player=self.player, desc=string.join(margs)))

	# Outgoing
	def send_event(self, event):
		try:
			self.events[event.evtype](event)
		except KeyError:
			pass

	def send_okay(self, event):
		self.send("OKAY cmd=%s" % event.name)
		if event.name == "JOIN":
			self.fieldstr = "GFLD"
	def send_error(self, event):
		self.send("%s cmd=%s %s" %
			  (event.evtype, event.name, event.desc))
	def send_clok(self, event):
		self.send("CLOK")
		self.send("BGML")
	def send_slct(self, event):
		self.send("EGML")
		self.send("SLCT")
	def send_jnok(self, event):
		self.send("JNOK g=%d p=%d name=%s" % (event.player.game.id, event.player.id, event.name))
	def send_pdon(self, event):
		self.send("PDON p=%d" % event.player.id)
	def send_gown(self, event):
		self.send("GOWN g=%d p=%d" % (event.game.id, event.player.id))
	def send_gplr(self, event):
		self.send("PLYR p=%d name=%s %s" % (event.player.id, event.name, event.desc))
	def send_pquit(self, event):
		self.send("KILL p=%d" % event.player.id)
	def send_winner(self, event):
		self.send("WINN p=%d" % event.player.id)
	def send_game(self, event):
		self.send("GAME g=%d owner=%d name=%s %s" % \
			  (event.game.id, event.player.id, event.name, event.desc))
	def send_year(self, event):
		self.send("YEAR y=%.3f" % event.quant)
		self.send("BRND")
	def send_time(self, event):
		self.send("TIME y=%.3f" % event.quant)
	def send_ernd(self, event):
		self.send("ERND")
	def send_psuicide(self, event):
		self.send("PKIL p=%d %s" % (event.player.id, event.desc))
	def send_c_or_g_fld(self, event):
		self.send("%s g=%d x=%d y=%d name=%s pl=%d" % \
			  (self.fieldstr, event.game.id, event.loc[0], \
			   event.loc[1], event.name, event.quant))
	def send_planet(self, event):
		self.send("PLNT i=%d owner=%d x=%d y=%d" % \
			  (event.planet.id, event.player.id, event.loc[0], event.loc[1]))
	def send_pindust(self, event):
		self.send("PIND i=%d indust=%d" % (event.planet.id, event.quant))
	def send_pships(self, event):
		self.send("PSHP i=%d %d=%d" % (event.planet.id, event.player.id, event.quant))
	def send_move_okay(self, event):
		self.send("OKAY cmd=MOVE")
		self.send("MVOK remain=%d eta=%.2f" % (event.quant, event.change))
	def send_message(self, event):
		self.send("SAID sayer=%d %s" % (event.player.id, event.desc))
	def send_battle(self, event):
		self.send("BATT i=%d att=%d def=%d surprise=%d" % \
			  (event.planet.id, event.player.id, event.tg_player.id, event.which))
		self.send("BSHP i=%d %d=%d %d=%d" % 
			  (event.planet.id, event.player.id, \
			   event.quant, event.tg_player.id, event.change))
	def send_attack(self, event):
		self.send("ATTK who=%d change=%d" % (event.which, event.change))
		self.send("BSHP i=%d %d=%d %d=%d" % 
			  (event.planet.id, event.player[0].id, \
			   event.quant[0], event.player[1].id, event.quant[1]))
	def send_bwin(self, event):
		self.send("BWIN i=%d winner=%d" % (event.planet.id, event.player.id))
	def send_fort(self, event):
		self.send("FORT i=%d new=%d %d=%d" % (event.planet.id, event.change, \
						      event.player.id, event.quant))

